/* 
 * File:   adcreader.c
 * Author: Carlo Alessandro Nicolau @ INFN
 * 
 * Firmware for BPD board.
 * ADC reader functions.
 * 
 */

#include "adcreader.h"
#include "configuration.h"
#include "hardware.h"
#include "generated/sources/variables.h"

// variables accessible from elsewhere
enum_averagingprescaler_t adcreader_averaging_prescaler = ADCREADER_AVERAGING_PRESCALER_DEFAULT;
analog_variable_sum_counter_t adcreader_averaging_prescaler_maxcount = (analog_variable_sum_counter_t)(1 << ADCREADER_AVERAGING_PRESCALER_DEFAULT);

// local variables
adcreader_state_t _state;   // ADC reader state machine status variable
uint8_t _idx = 0;           // next analog variable index to acquire
uint8_t _state_timeout;     // timeout counter used for state transitions

// local functions
void _sample_acquired(void);
void _update_alarm(analogalarm_info_t* _p_alarm, analog_value_t _value);
bool _check_alarm_firing(analogalarm_info_t* _p_alarm, alarm_callback_t _callback);
void _update_averaging(analog_variable_t* _p_analog_variable);

void adcreader_reset(void) {    
    _state = ADCREADER_RESET;
    _idx = 0;
    _state_timeout = ADCREADER_INIT_TIMEOUT+1;
}

void adcreader_start(void) {
    _state = ADCREADER_MUXCHANGE;
}

void adcreader_tick_irq(void) {
    // this function is called from IRQ
    
    // if in reset state, do nothing
    if(_state == ADCREADER_RESET) return;
    
    // decrement the next step timeout counter
    _state_timeout--;
    
    // exit if timeout was not reached
    if(_state_timeout > 0) return;
    
    // execute next step
    switch(_state) {
        case ADCREADER_MUXCHANGE:            
            // check if the adc has finished last conversion
            if(!adc_conversion_done()) {
                // conversion not done, skip this cycle (will retry next time)
                _state_timeout++;
                return;
            }            
            // stores the adc reading in the proper slot
            analog_value_t adcvalue = adc_getvalue();            
            analog_variables[_idx].value = adcvalue;
            // check alarms for current measurement
            _sample_acquired();            
            // increment (wrapping) the current read channel            
            _idx = (_idx + 1) % ANALOG_VARIABLES_COUNT;
            // switch the adc mux to the next channel
            adc_select_channel(analog_variables[_idx].channel);
            // update current state
            _state = ADCREADER_CONVSTART;                        
            // update current state timeout
            _state_timeout = ADCREADER_CONVSTART_TIMEOUT;                        
            break;
        case ADCREADER_CONVSTART:            
            // start the adc conversion
            adc_start_conversion();
            // update current state
            _state = ADCREADER_MUXCHANGE;            
            // update current state timeout
            _state_timeout = ADCREADER_MUXCHANGE_TIMEOUT;                
            break;
        default:
            break;
    }
}

void _sample_acquired(void) { 
    // store current analog variable pointer
    analog_variable_t *analog_variable = &(analog_variables[_idx]);
    
    // update analog variable max value
    if(analog_variable->value_max < analog_variable->value) {    
        analog_variable->value_max = analog_variable->value;
    }    
        
    // manage fast and slow alarms (increment timeout counter if threshold exceeded, else reset to zero)
    _update_alarm(&(analog_variable->alarms.fast), analog_variable->value);
    _update_alarm(&(analog_variable->alarms.slow), analog_variable->value);    
    // check timeout on fast alarm, and fire in case
    if(false == _check_alarm_firing(&(analog_variable->alarms.fast), analog_variable->alarms.callback)) {
        // fast alarm did not fire, check also slow alarm
        _check_alarm_firing(&(analog_variable->alarms.slow), analog_variable->alarms.callback);
    }
    _update_averaging(analog_variable);
}

void _update_alarm(analogalarm_info_t* _p_alarm, analog_value_t _value) {
    // increment timeout counter if threshold exceeded, else reset to zero
    if(_value > _p_alarm->threshold) {
        // increment counter
        _p_alarm->timeout_counter += 1;
        // avoid wrapping on increment (not very elegant...)
        if(_p_alarm->timeout_counter == 0) {
            _p_alarm->timeout_counter -= 1;
        }
        // update alarm counter stats
        if(_p_alarm->timeout_counter_max < _p_alarm->timeout_counter) {
            _p_alarm->timeout_counter_max = _p_alarm->timeout_counter;
        }
    }else{
        // or reset the counter
        _p_alarm->timeout_counter = 0;
    }
}

bool _check_alarm_firing(analogalarm_info_t* _p_alarm, alarm_callback_t _callback) {
    // return true if the alarm has fired        
        
    if(_p_alarm->enabled) {
        // check if timeout was reached
        if(_p_alarm->timeout_counter > _p_alarm->timeout) {
            // perform the alarm action
            _callback();
            // increment the fire counter
            _p_alarm->firecount += 1;
            // avoid wrapping on increment (not very elegant...)
            if(_p_alarm->firecount == 0) {
                _p_alarm->firecount -= 1;
            }
            // reset alarm timeout counter
            _p_alarm->timeout_counter = 0;
            // remember for next alarm
            return true;
        }
    }
    return false;
}

void _update_averaging(analog_variable_t* _p_analog_variable) {
    _p_analog_variable->sum_curr += _p_analog_variable->value;
    _p_analog_variable->sum_counter++;
    if(_p_analog_variable->sum_counter >= adcreader_averaging_prescaler_maxcount) {
        _p_analog_variable->sum_last = _p_analog_variable->sum_curr;
        _p_analog_variable->sum_curr = 0;
        _p_analog_variable->sum_counter = 0;
    }    
}

