/*
 * File:   communication.c
 * Author: Carlo Alessandro Nicolau @ INFN
 * 
 * Firmware for BPD board.
 * Communication functions.
 * 
 */

#include "communication.h"
#include "configuration.h"
#include "errorcodes.h"
#include "typedefs.h"
#include "errors.h"
#include "commands.h"
#include "hardware.h"
#include "generated/sources/commandstable.h"
#include "generated/sources/variables.h" // TODO remove this include

// local ("private") variables
communication_parser_state_t _parser_status;    // communication parser current status
command_code_t _parser_next_cmdcode;            // command code of the next command
command_datalen_t _parser_data_downcounter;     // packet payload data length down-counter
uint8_t _databuf[COMMUNICATION_DATABUF_SIZE];   // packet payload data buffer
command_datalen_t _databuf_index;               // packet payload data buffer index (used by payload data retriever and appender also)
uint8_t _parser_checksum;                       // running checksum for the received data
// variables used by payload data retriever and appender
// command_datalen_t _payload_index;

// local ("private") functions
void _checksum_update(uint8_t _byte);
void _checksum_reset(void);
uint8_t _checksum_value(void);
bool _inrange(uint8_t _value, uint8_t _min, uint8_t _max);
void _communication_error(err_t _err);
err_t _ascii2hex(uint8_t _byte, uint8_t* _p_nibble);
uint8_t _hex2ascii(uint8_t _nibble);
void _serial_checksum_update(uint8_t _byte);

void communication_parser_reset(void) {    
    // initialize the state machine status
    _parser_status = WAITING_SOF;    
    // invalidate next command code
    _parser_next_cmdcode = CMDCODE_INVALID_REQ;
    // initialize the number of payload data bytes to receive
    _parser_data_downcounter = 0;    
    // reset the receive payload data queue
    _databuf_index = 0;
}

void communication_parser_appendbyte(uint8_t _byte) {
    // check if a previously received command was not managed yet and, if so,
    // notify the error and reset the receiver
    if(communication_parser_commandready()) {
        error_notify(ERR_COMMUNICATION_PENDINGCMD);
    }    
    // update the checksum calculator
    // a note here: the following call is here to preserve some code space but
    // it would more logically placed inside the relevant switch-case cases;
    // notice also that the checksum value is reset on receiving a SOF byte
    _checksum_update(_byte);    
    // some space allocated for temporary data    
    command_info_t* tmp_p_cmdinfo;
    static uint8_t tmp_nibble_even; // these value must be remembered between calls
    static uint8_t tmp_nibble_odd; // these value must be remembered between calls
    uint8_t tmp_new_byte;
    err_t retval;    
    // switch-case for the communication parser state machine
    switch(_parser_status) {
        case WAITING_SOF:             
            // the received byte should be a start-of-frame, check
            if(_byte != COMMUNICATION_PACKET_SOF) {
                // it's not: notify error, reset receiver and exit
                _communication_error(ERR_COMMUNICATION_WRONG_SOF);
                return;
            }
            // initialize checksum calculator
            _checksum_reset();            
            // update next state
            _parser_status = WAITING_ADDR;
            break;
        case WAITING_ADDR:            
            // the received byte should be the address, check
            if(_byte != COMMUNICATION_PACKET_ADDR) {
                // it's not: notify error, reset receiver and exit
                _communication_error(ERR_COMMUNICATION_WRONG_ADDR);
                return;
            }                       
            // update next state
            _parser_status = WAITING_CMD;
            break;
        case WAITING_CMD:            
            // the received byte should be the command code,
            // check if the command code exists and retrieve its info
            retval = commands_retrieve_info_bycode((command_code_t)_byte, &tmp_p_cmdinfo);
            if(retval != NOERR) {
                // it's not valid: notify error, reset receiver and exit
                _communication_error(retval);
                return;
            }
            // the command code is valid
            // set the command code of the next command to be executed and set
            // the  number of data to receive for the command to complete
            _parser_next_cmdcode = (command_code_t)_byte;
            _parser_data_downcounter = tmp_p_cmdinfo->request_payload_len;
            // update next state
            if(_parser_data_downcounter > 0) {            
                _parser_status = WAITING_DATA_EVEN;
                // next states will populate communication_databuf[]
                _databuf_index = 0;
            }else{          
                _parser_status = WAITING_CHK;                
            }
            break;
        case WAITING_DATA_EVEN:            
            // the received byte should be an even payload data nibble
            // check that the received byte is a valid hex code, and convert it to a 4-bit value
            retval = _ascii2hex(_byte, &tmp_nibble_even);
            if(retval != NOERR) {
                // notify error, reset receiver and exit
                _communication_error(retval);
                return;
            }
            // update next state
            _parser_status = WAITING_DATA_ODD;
            break;
        case WAITING_DATA_ODD:            
            // the received byte should be an odd payload data nibble
            // check that the received byte is a valid hex code, and convert it to a 4-bit value
            retval = _ascii2hex(_byte, &tmp_nibble_odd);
            if(retval != NOERR) {
                // notify error and reset receiver
                _communication_error(retval);
                return;
            }
            // join the two received nibbles into one byte
            tmp_new_byte = (tmp_nibble_even << 4) + tmp_nibble_odd;
            // check if the payload data receive buffer is full
            if(_databuf_index >= COMMUNICATION_DATABUF_SIZE) {                
                // notify error, reset receiver and exit
                _communication_error(ERR_COMMUNICATION_RECVBUFF_FULL);
                return;
            }
            // store data in the payload data receive buffer and increment pointer
            _databuf[_databuf_index++] = tmp_new_byte;
            // decrement number of payload data left to receive
            _parser_data_downcounter--;
            // update next state
            if(_parser_data_downcounter == 0) {
                // next byte is check code
                _parser_status = WAITING_CHK;
            }else{
                // next byte is a data word 
                _parser_status = WAITING_DATA_EVEN;
            }            
            break;
        case WAITING_CHK:            
            // the received byte should be the check byte
            // so (because it has already been computed) the checksum should be zero
            if(_checksum_value() != 0) {
                // checksum error: notify, reset receiver and exit
                _communication_error(ERR_COMMUNICATION_CHECKSUM);
                return;
            }            
            // update next state
            // if the checksum received byte was equal to the end of frame value,
            // then the packet is completed, otherwise wait for end of frame
            if(_byte == COMMUNICATION_PACKET_EOF) {
                _parser_status = COMMAND_READY;
            }else{
                _parser_status = WAITING_EOF;                
            }
            break;
        case WAITING_EOF:            
            // the received byte should be the end-of-frame byte, check
            if(_byte != COMMUNICATION_PACKET_EOF) {
                // it does not match: notify, reset receiver and exit
                _communication_error(ERR_COMMUNICATION_WRONG_EOF);
                return;
            }            
            // update next state
            _parser_status = COMMAND_READY;
            break;
        default:                 
            // unknown condition: notify and reset
            _communication_error(ERR_COMMUNICATION_UNKNWOWN);
            break;        
    }  
}

bool communication_parser_commandready(void) {
    // return true if a new command can be executed
    return (_parser_status == COMMAND_READY);
}

err_t communication_parser_executecommand(void) {
    // This function assumes the following conditions are verified:
    //  - communication_parser_next_cmdcode contains the command code to be executed
    //  - communication_databuf contains (from location [0]) the payload data
    //  - communication_databuf_index contains the number of payload data received
    
    // retrieve info about the command (it's a double check)
    command_info_t* p_cmd_info;
    err_t retval;
    retval = commands_retrieve_info_bycode(_parser_next_cmdcode, &p_cmd_info);
    if(retval != NOERR) {        
        return ERR_COMMUNICATION_EXEC;
    }
    
    // verify data payload length
    if(p_cmd_info->request_payload_len != _databuf_index) {
        return ERR_COMMUNICATION_EXEC;
    }
    
    // reset the request payload retriever
    communication_request_payload_retriever_reset();
    // execute the command callback function
    return p_cmd_info->callback_ptr();
}

void communication_transmitter_sendresponse(void) {
    // This function assumes the following conditions are verified:   
    // - the communication_parser_next_cmdcode contains the command code that was just served
    // - the response payload data is into the buffer comm_databuf[]
    // - the number of payload bytes is into _databuf_index
    
    // retrieve information about the executed command
    err_t ret;
    command_info_t* p_cmdinfo;
    ret = commands_retrieve_info_bycode(_parser_next_cmdcode, &p_cmdinfo);
    // ret is ignored: we assume no problems here
    // get the response code to use
    command_code_t resp_code = p_cmdinfo->response_code;
    
    // send start of frame
    serialport_send(COMMUNICATION_PACKET_SOF);
    
    // init checksum calculator
    _checksum_reset();            
    
    // send address field and update checksum
    _serial_checksum_update(COMMUNICATION_PACKET_ADDR);
    
    // send command response code and update checksum
    _serial_checksum_update(resp_code);
    
    // send the data payload
    command_datalen_t idx;
    for(idx = 0; idx < _databuf_index; idx++) {
        uint8_t byte = _databuf[idx];
        uint8_t ascii_high = _hex2ascii((byte >> 4) & 0x0f);
        uint8_t ascii_low = _hex2ascii(byte & 0x0f);        
    
        // send high nibble and update checksum
        _serial_checksum_update(ascii_high);
    
        // send low nibble and update checksum
        _serial_checksum_update(ascii_low);
    }
    
    // send the checksum
    serialport_send(_checksum_value());
    
    // if checksum != PACKET_EOF, send PACKET_EOF
    serialport_send(COMMUNICATION_PACKET_EOF);
}

void communication_request_payload_retriever_reset(void) {    
    _databuf_index = 0;
}

uint8_t communication_request_payload_retrieve_uint8_t(void) {
    uint8_t retval = _databuf[_databuf_index];
    _databuf_index += 1;
    return retval;
}

uint16_t communication_request_payload_retrieve_uint16_t(void) {    
    uint16_t retval = ((uint16_t)_databuf[_databuf_index] << 8) + _databuf[_databuf_index+1];
    _databuf_index += 2;
    return retval;
}

void communication_response_payload_appender_reset(void) {
    _databuf_index = 0;
}

void communication_response_payload_append_uint8_t(uint8_t _value) {    
    _databuf[_databuf_index] = _value;
    _databuf_index += 1;
}

void communication_response_payload_append_uint16_t(uint16_t _value) {
    _databuf[_databuf_index] = (uint8_t)((_value >> 8) & 0xff);
    _databuf[_databuf_index+1] = (uint8_t)(_value & 0xff);   
    _databuf_index += 2;
}

void communication_response_payload_append_uint32_t(uint32_t _value) {
    _databuf[_databuf_index] = (uint8_t)((_value >> 24) & 0xff);
    _databuf[_databuf_index+1] = (uint8_t)((_value >> 16) & 0xff);
    _databuf[_databuf_index+2] = (uint8_t)((_value >> 8) & 0xff);
    _databuf[_databuf_index+3] = (uint8_t)(_value & 0xff);   
    _databuf_index += 4;
}

void _checksum_update(uint8_t _byte) {
    _parser_checksum -= _byte;
}

void _checksum_reset(void) {
    _parser_checksum = 0;
}

uint8_t _checksum_value(void) {
    return _parser_checksum;
}

bool _inrange(uint8_t _value, uint8_t _min, uint8_t _max) {
    if(_value > _max) return false;
    if(_value < _min) return false;
    return true;    
}

void _communication_error(err_t _err) {
    error_notify(_err);
    communication_parser_reset();    
}

err_t _ascii2hex(uint8_t _byte, uint8_t* _p_nibble) {
    // correct values are '0'...'9', 'a'...'f', 'A'...'F'
    if(_inrange(_byte, '0', '9')) {
        (*_p_nibble) = (uint8_t)(_byte - '0');
    }else if(_inrange(_byte, 'a', 'f')) {
        (*_p_nibble) = (uint8_t)(_byte - 'a' + 10);
    }else if(_inrange(_byte, 'A', 'F')) {
        // MAYBE: this is not necessary
        (*_p_nibble) = (uint8_t)(_byte - 'A' + 10);
    }else {
        return ERR_COMMUNICATION_ASCII2HEX;
    }       
    return NOERR;    
}

uint8_t _hex2ascii(uint8_t _nibble) {
    _nibble = _nibble & 0x0f;
    if(_nibble < 0x0a) {
        return '0' + _nibble;
    }
    return 'a' + _nibble - 10;
}

void _serial_checksum_update(uint8_t _byte) {
    serialport_send(_byte);
    _checksum_update(_byte);    
}